'use strict';

var _path = require('path');
var sander = require('sander');
var MagicString = require('magic-string');
MagicString = ('default' in MagicString ? MagicString['default'] : MagicString);
var acorn = require('acorn');

var keys = Object.keys;

var hasOwnProp = Object.prototype.hasOwnProperty;function has(obj, prop) {
	return hasOwnProp.call(obj, prop);
}

function sequence(arr, callback) {
	var len = arr.length;
	var results = new Array(len);

	var promise = sander.Promise.resolve();

	function next(i) {
		return promise.then(function () {
			return callback(arr[i], i);
		}).then(function (result) {
			return results[i] = result;
		});
	}

	var i = undefined;

	for (i = 0; i < len; i += 1) {
		promise = next(i);
	}

	return promise.then(function () {
		return results;
	});
}

var shouldSkip = undefined;
var shouldAbort = undefined;
function walk(ast, _ref) {
	var enter = _ref.enter;
	var leave = _ref.leave;

	shouldAbort = false;
	visit(ast, null, enter, leave);
}

var context = {
	skip: function () {
		return shouldSkip = true;
	},
	abort: function () {
		return shouldAbort = true;
	}
};

var childKeys = {};

var toString = Object.prototype.toString;

function isArray(thing) {
	return toString.call(thing) === '[object Array]';
}

function visit(node, parent, enter, leave) {
	if (!node || shouldAbort) return;

	if (enter) {
		shouldSkip = false;
		enter.call(context, node, parent);
		if (shouldSkip || shouldAbort) return;
	}

	var keys = childKeys[node.type] || (childKeys[node.type] = Object.keys(node).filter(function (key) {
		return typeof node[key] === 'object';
	}));

	var key = undefined,
	    value = undefined,
	    i = undefined,
	    j = undefined;

	i = keys.length;
	while (i--) {
		key = keys[i];
		value = node[key];

		if (isArray(value)) {
			j = value.length;
			while (j--) {
				visit(value[j], node, enter, leave);
			}
		} else if (value && value.type) {
			visit(value, node, enter, leave);
		}
	}

	if (leave && !shouldAbort) {
		leave(node, parent);
	}
}

function Scope___classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Scope = (function () {
	function Scope(options) {
		Scope___classCallCheck(this, Scope);

		options = options || {};

		this.parent = options.parent;
		this.depth = this.parent ? this.parent.depth + 1 : 0;
		this.names = options.params || [];
		this.isBlockScope = !!options.block;
	}

	Scope.prototype.add = function add(name, isBlockDeclaration) {
		if (!isBlockDeclaration && this.isBlockScope) {
			// it's a `var` or function declaration, and this
			// is a block scope, so we need to go up
			this.parent.add(name, isBlockDeclaration);
		} else {
			this.names.push(name);
		}
	};

	Scope.prototype.contains = function contains(name) {
		return !!this.findDefiningScope(name);
	};

	Scope.prototype.findDefiningScope = function findDefiningScope(name) {
		if (~this.names.indexOf(name)) {
			return this;
		}

		if (this.parent) {
			return this.parent.findDefiningScope(name);
		}

		return null;
	};

	return Scope;
})();

function getName(x) {
	return x.name;
}

function quoteId(x) {
	return "'" + x.id + "'";
}

function req(x) {
	return "require('" + x.id + "')";
}

function getLocation(source, charIndex) {
	var lines = source.split('\n');
	var len = lines.length;

	var lineStart = 0;
	var i = undefined;

	for (i = 0; i < len; i += 1) {
		var line = lines[i];
		var lineEnd = lineStart + line.length + 1; // +1 for newline

		if (lineEnd > charIndex) {
			return { line: i + 1, column: charIndex - lineStart };
		}

		lineStart = lineEnd;
	}

	throw new Error('Could not determine location of character');
}

var _analyse = analyse;
function analyse(ast, magicString, module) {
	var scope = new Scope();
	var currentTopLevelStatement = undefined;

	function addToScope(declarator) {
		var name = declarator.id.name;
		scope.add(name, false);

		if (!scope.parent) {
			currentTopLevelStatement._defines[name] = true;
		}
	}

	function addToBlockScope(declarator) {
		var name = declarator.id.name;
		scope.add(name, true);

		if (!scope.parent) {
			currentTopLevelStatement._defines[name] = true;
		}
	}

	// first we need to generate comprehensive scope info
	var previousStatement = null;
	var commentIndex = 0;

	ast.body.forEach(function (statement) {
		currentTopLevelStatement = statement; // so we can attach scoping info

		Object.defineProperties(statement, {
			_defines: { value: {} },
			_modifies: { value: {} },
			_dependsOn: { value: {} },
			_included: { value: false, writable: true },
			_module: { value: module },
			_source: { value: magicString.snip(statement.start, statement.end) }, // TODO don't use snip, it's a waste of memory
			_margin: { value: [0, 0] },
			_leadingComments: { value: [] },
			_trailingComment: { value: null, writable: true } });

		var trailing = !!previousStatement;

		// attach leading comment
		do {
			var comment = module.comments[commentIndex];

			if (!comment || comment.end > statement.start) break;

			// attach any trailing comment to the previous statement
			if (trailing && !/\n/.test(magicString.slice(previousStatement.end, comment.start))) {
				previousStatement._trailingComment = comment;
			}

			// then attach leading comments to this statement
			else {
				statement._leadingComments.push(comment);
			}

			commentIndex += 1;
			trailing = false;
		} while (module.comments[commentIndex]);

		// determine margin
		var previousEnd = previousStatement ? (previousStatement._trailingComment || previousStatement).end : 0;
		var start = (statement._leadingComments[0] || statement).start;

		var gap = magicString.original.slice(previousEnd, start);
		var margin = gap.split('\n').length;

		if (previousStatement) previousStatement._margin[1] = margin;
		statement._margin[0] = margin;

		walk(statement, {
			enter: function (node) {
				var newScope = undefined;

				magicString.addSourcemapLocation(node.start);

				switch (node.type) {
					case 'FunctionExpression':
					case 'FunctionDeclaration':
					case 'ArrowFunctionExpression':
						var names = node.params.map(getName);

						if (node.type === 'FunctionDeclaration') {
							addToScope(node);
						} else if (node.type === 'FunctionExpression' && node.id) {
							names.push(node.id.name);
						}

						newScope = new Scope({
							parent: scope,
							params: names, // TODO rest params?
							block: false
						});

						break;

					case 'BlockStatement':
						newScope = new Scope({
							parent: scope,
							block: true
						});

						break;

					case 'CatchClause':
						newScope = new Scope({
							parent: scope,
							params: [node.param.name],
							block: true
						});

						break;

					case 'VariableDeclaration':
						node.declarations.forEach(node.kind === 'let' ? addToBlockScope : addToScope); // TODO const?
						break;

					case 'ClassDeclaration':
						addToScope(node);
						break;
				}

				if (newScope) {
					Object.defineProperty(node, '_scope', { value: newScope });
					scope = newScope;
				}
			},
			leave: function (node) {
				if (node === currentTopLevelStatement) {
					currentTopLevelStatement = null;
				}

				if (node._scope) {
					scope = scope.parent;
				}
			}
		});

		previousStatement = statement;
	});

	// then, we need to find which top-level dependencies this statement has,
	// and which it potentially modifies
	ast.body.forEach(function (statement) {
		function checkForReads(node, parent) {
			if (node.type === 'Identifier') {
				// disregard the `bar` in `foo.bar` - these appear as Identifier nodes
				if (parent.type === 'MemberExpression' && node !== parent.object) {
					return;
				}

				// disregard the `bar` in { bar: foo }
				if (parent.type === 'Property' && node !== parent.value) {
					return;
				}

				var definingScope = scope.findDefiningScope(node.name);

				if ((!definingScope || definingScope.depth === 0) && !statement._defines[node.name]) {
					statement._dependsOn[node.name] = true;
				}
			}
		}

		function checkForWrites(node) {
			function addNode(node, disallowImportReassignments) {
				while (node.type === 'MemberExpression') {
					node = node.object;
				}

				// disallow assignments/updates to imported bindings and namespaces
				if (disallowImportReassignments && has(module.imports, node.name) && !scope.contains(node.name)) {
					var err = new Error('Illegal reassignment to import \'' + node.name + '\'');
					err.file = module.path;
					err.loc = getLocation(module.code.toString(), node.start);
					throw err;
				}

				if (node.type !== 'Identifier') {
					return;
				}

				statement._modifies[node.name] = true;
			}

			if (node.type === 'AssignmentExpression') {
				addNode(node.left, true);
			} else if (node.type === 'UpdateExpression') {
				addNode(node.argument, true);
			} else if (node.type === 'CallExpression') {
				node.arguments.forEach(function (arg) {
					return addNode(arg, false);
				});
			}

			// TODO UpdateExpressions, method calls?
		}

		walk(statement, {
			enter: function (node, parent) {
				// skip imports
				if (/^Import/.test(node.type)) return this.skip();

				if (node._scope) scope = node._scope;

				checkForReads(node, parent);
				checkForWrites(node, parent);

				//if ( node.type === 'ReturnStatement')
			},
			leave: function (node) {
				if (node._scope) scope = scope.parent;
			}
		});
	});

	ast._scope = scope;
}

var reservedWords = 'break case class catch const continue debugger default delete do else export extends finally for function if import in instanceof let new return super switch this throw try typeof var void while with yield enum await implements package protected static interface private public'.split(' ');
var builtins = 'Infinity NaN undefined null true false eval uneval isFinite isNaN parseFloat parseInt decodeURI decodeURIComponent encodeURI encodeURIComponent escape unescape Object Function Boolean Symbol Error EvalError InternalError RangeError ReferenceError SyntaxError TypeError URIError Number Math Date String RegExp Array Int8Array Uint8Array Uint8ClampedArray Int16Array Uint16Array Int32Array Uint32Array Float32Array Float64Array Map Set WeakMap WeakSet SIMD ArrayBuffer DataView JSON Promise Generator GeneratorFunction Reflect Proxy Intl'.split(' ');

var blacklisted = {};
reservedWords.concat(builtins).forEach(function (word) {
	return blacklisted[word] = true;
});
function makeLegalIdentifier(str) {
	str = str.replace(/[^$_a-zA-Z0-9]/g, '_');
	if (/\d/.test(str[0]) || blacklisted[str]) str = '_' + str;

	return str;
}

function Module___classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var emptyArrayPromise = sander.Promise.resolve([]);

var Module = (function () {
	function Module(_ref) {
		var _this = this;

		var path = _ref.path;
		var code = _ref.code;
		var bundle = _ref.bundle;

		Module___classCallCheck(this, Module);

		this.bundle = bundle;
		this.path = path;
		this.relativePath = _path.relative(bundle.base, path).slice(0, -3); // remove .js

		this.code = new MagicString(code, {
			filename: path
		});

		this.suggestedNames = {};
		this.comments = [];

		try {
			this.ast = acorn.parse(code, {
				ecmaVersion: 6,
				sourceType: 'module',
				onComment: function (block, text, start, end) {
					return _this.comments.push({ block: block, text: text, start: start, end: end });
				}
			});
		} catch (err) {
			err.file = path;
			throw err;
		}

		this.analyse();
	}

	Module.prototype.analyse = function analyse() {
		var _this2 = this;

		// imports and exports, indexed by ID
		this.imports = {};
		this.exports = {};

		var commentIndex = 0;
		var previousNode = undefined;
		var previousEnd = 0;

		this.ast.body.forEach(function (node) {
			var source = undefined;

			// import foo from './foo';
			// import { bar } from './bar';
			if (node.type === 'ImportDeclaration') {
				source = node.source.value;

				node.specifiers.forEach(function (specifier) {
					var isDefault = specifier.type === 'ImportDefaultSpecifier';
					var isNamespace = specifier.type === 'ImportNamespaceSpecifier';

					var localName = specifier.local.name;
					var name = isDefault ? 'default' : isNamespace ? '*' : specifier.imported.name;

					if (has(_this2.imports, localName)) {
						var err = new Error('Duplicated import \'' + localName + '\'');
						err.file = _this2.path;
						err.loc = getLocation(_this2.code.original, specifier.start);
						throw err;
					}

					_this2.imports[localName] = {
						source: source,
						name: name,
						localName: localName
					};
				});
			} else if (/^Export/.test(node.type)) {
				// export default function foo () {}
				// export default foo;
				// export default 42;
				if (node.type === 'ExportDefaultDeclaration') {
					var isDeclaration = /Declaration$/.test(node.declaration.type);

					_this2.exports.default = {
						node: node,
						name: 'default',
						localName: isDeclaration ? node.declaration.id.name : 'default',
						isDeclaration: isDeclaration
					};
				}

				// export { foo, bar, baz }
				// export var foo = 42;
				// export function foo () {}
				else if (node.type === 'ExportNamedDeclaration') {
					// export { foo } from './foo';
					source = node.source && node.source.value;

					if (node.specifiers.length) {
						// export { foo, bar, baz }
						node.specifiers.forEach(function (specifier) {
							var localName = specifier.local.name;
							var exportedName = specifier.exported.name;

							_this2.exports[exportedName] = {
								localName: localName,
								exportedName: exportedName
							};

							if (source) {
								_this2.imports[localName] = {
									source: source,
									localName: localName,
									name: exportedName
								};
							}
						});
					} else {
						var declaration = node.declaration;

						var _name = undefined;

						if (declaration.type === 'VariableDeclaration') {
							// export var foo = 42
							_name = declaration.declarations[0].id.name;
						} else {
							// export function foo () {}
							_name = declaration.id.name;
						}

						_this2.exports[_name] = {
							node: node,
							localName: _name,
							expression: declaration
						};
					}
				}
			}
		});

		_analyse(this.ast, this.code, this);

		this.definedNames = this.ast._scope.names.slice();

		this.canonicalNames = {};

		this.definitions = {};
		this.definitionPromises = {};
		this.modifications = {};

		this.ast.body.forEach(function (statement) {
			Object.keys(statement._defines).forEach(function (name) {
				_this2.definitions[name] = statement;
			});

			Object.keys(statement._modifies).forEach(function (name) {
				if (!has(_this2.modifications, name)) {
					_this2.modifications[name] = [];
				}

				_this2.modifications[name].push(statement);
			});
		});
	};

	Module.prototype.getCanonicalName = function getCanonicalName(localName) {
		if (has(this.suggestedNames, localName)) {
			localName = this.suggestedNames[localName];
		}

		if (!has(this.canonicalNames, localName)) {
			var canonicalName = undefined;

			if (has(this.imports, localName)) {
				var importDeclaration = this.imports[localName];
				var _module = importDeclaration.module;

				if (importDeclaration.name === '*') {
					canonicalName = _module.suggestedNames['*'];
				} else {
					var exporterLocalName = undefined;

					if (_module.isExternal) {
						exporterLocalName = importDeclaration.name;
					} else {
						var exportDeclaration = _module.exports[importDeclaration.name];
						exporterLocalName = exportDeclaration.localName;
					}

					canonicalName = _module.getCanonicalName(exporterLocalName);
				}
			} else {
				canonicalName = localName;
			}

			this.canonicalNames[localName] = canonicalName;
		}

		return this.canonicalNames[localName];
	};

	Module.prototype.define = function define(name) {
		var _this3 = this;

		// shortcut cycles. TODO this won't work everywhere...
		if (has(this.definitionPromises, name)) {
			return emptyArrayPromise;
		}

		var promise = undefined;

		// The definition for this name is in a different module
		if (has(this.imports, name)) {
			(function () {
				var importDeclaration = _this3.imports[name];

				promise = _this3.bundle.fetchModule(importDeclaration.source, _this3.path).then(function (module) {
					importDeclaration.module = module;

					if (importDeclaration.name === 'default') {
						// TODO this seems ropey
						var localName = importDeclaration.localName;
						var suggestion = has(_this3.suggestedNames, localName) ? _this3.suggestedNames[localName] : localName;
						module.suggestName('default', suggestion);
					}

					if (module.isExternal) {
						if (importDeclaration.name === 'default') {
							module.needsDefault = true;
						} else {
							module.needsNamed = true;
						}

						module.importedByBundle.push(importDeclaration);
						return emptyArrayPromise;
					}

					if (importDeclaration.name === '*') {
						module.suggestName('*', importDeclaration.localName);
						module.suggestName('default', '' + importDeclaration.localName + '__default');

						// we need to create an internal namespace
						if (! ~_this3.bundle.internalNamespaceModules.indexOf(module)) {
							_this3.bundle.internalNamespaceModules.push(module);
						}

						return module.expandAllStatements();
					}

					var exportDeclaration = module.exports[importDeclaration.name];

					if (!exportDeclaration) {
						throw new Error('Module ' + module.path + ' does not export ' + importDeclaration.name + ' (imported by ' + _this3.path + ')');
					}

					return module.define(exportDeclaration.localName);
				});
			})();
		}

		// The definition is in this module
		else if (name === 'default' && this.exports.default.isDeclaration) {
			// We have something like `export default foo` - so we just start again,
			// searching for `foo` instead of default
			promise = this.define(this.exports.default.name);
		} else {
			var statement = undefined;

			if (name === 'default') {
				// TODO can we use this.definitions[ name ], as below?
				statement = this.exports.default.node;
			} else {
				statement = this.definitions[name];
			}

			if (statement && !statement._included) {
				promise = this.expandStatement(statement);
			}
		}

		this.definitionPromises[name] = promise || emptyArrayPromise;
		return this.definitionPromises[name];
	};

	Module.prototype.expandStatement = function expandStatement(statement) {
		var _this4 = this;

		if (statement._included) return emptyArrayPromise;
		statement._included = true;

		var result = [];

		// We have a statement, and it hasn't been included yet. First, include
		// the statements it depends on
		var dependencies = Object.keys(statement._dependsOn);

		return sequence(dependencies, function (name) {
			return _this4.define(name).then(function (definition) {
				result.push.apply(result, definition);
			});
		})

		// then include the statement itself
		.then(function () {
			result.push(statement);
		})

		// then include any statements that could modify the
		// thing(s) this statement defines
		.then(function () {
			return sequence(keys(statement._defines), function (name) {
				var modifications = has(_this4.modifications, name) && _this4.modifications[name];

				if (modifications) {
					return sequence(modifications, function (statement) {
						if (!statement._included) {
							return _this4.expandStatement(statement).then(function (statements) {
								result.push.apply(result, statements);
							});
						}
					});
				}
			});
		})

		// the `result` is an array of statements needed to define `name`
		.then(function () {
			return result;
		});
	};

	Module.prototype.expandAllStatements = function expandAllStatements(isEntryModule) {
		var _this5 = this;

		var allStatements = [];

		return sequence(this.ast.body, function (statement) {
			// skip already-included statements
			if (statement._included) return;

			// skip import declarations
			if (statement.type === 'ImportDeclaration') {
				// unless they're empty, in which case assume we're importing them for the side-effects
				// THIS IS NOT FOOLPROOF. Probably need /*rollup: include */ or similar
				if (!statement.specifiers.length) {
					return _this5.bundle.fetchModule(statement.source.value, _this5.path).then(function (module) {
						statement.module = module;
						return module.expandAllStatements();
					}).then(function (statements) {
						allStatements.push.apply(allStatements, statements);
					});
				}

				return;
			}

			// skip `export { foo, bar, baz }`
			if (statement.type === 'ExportNamedDeclaration' && statement.specifiers.length) {
				// but ensure they are defined, if this is the entry module
				if (isEntryModule) {
					return _this5.expandStatement(statement).then(function (statements) {
						allStatements.push.apply(allStatements, statements);
					});
				}

				return;
			}

			// include everything else
			return _this5.expandStatement(statement).then(function (statements) {
				allStatements.push.apply(allStatements, statements);
			});
		}).then(function () {
			return allStatements;
		});
	};

	Module.prototype.rename = function rename(name, replacement) {
		this.canonicalNames[name] = replacement;
	};

	Module.prototype.suggestName = function suggestName(exportName, suggestion) {
		if (!this.suggestedNames[exportName]) {
			this.suggestedNames[exportName] = makeLegalIdentifier(suggestion);
		}
	};

	return Module;
})();

function ExternalModule___classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var ExternalModule = (function () {
	function ExternalModule(id) {
		ExternalModule___classCallCheck(this, ExternalModule);

		this.id = id;
		this.name = null;
		this.module = null;

		this.isExternal = true;
		this.importedByBundle = [];

		this.canonicalNames = {};
		this.suggestedNames = {};

		this.needsDefault = null;
		this.needsNamed = null;
	}

	ExternalModule.prototype.getCanonicalName = function getCanonicalName(name) {
		if (name === 'default') {
			return this.needsNamed ? '' + this.name + '__default' : this.name;
		}

		if (name === '*') {
			return this.name;
		}

		// TODO this depends on the output format... works for CJS etc but not ES6
		return '' + this.name + '.' + name;
	};

	ExternalModule.prototype.rename = function rename(name, replacement) {
		this.canonicalNames[name] = replacement;
	};

	ExternalModule.prototype.suggestName = function suggestName(exportName, suggestion) {
		if (!this.suggestedNames[exportName]) {
			this.suggestedNames[exportName] = suggestion;
		}
	};

	return ExternalModule;
})();

function amd(bundle, magicString, exportMode, options) {
	var indentStr = magicString.getIndentString();

	var deps = bundle.externalModules.map(quoteId);
	var args = bundle.externalModules.map(getName);

	if (exportMode === 'named') {
		args.unshift('exports');
		deps.unshift('\'exports\'');
	}

	var params = (has(options, 'moduleId') ? '[\'' + options.moduleId + '\'], ' : '') + (deps.length ? '[' + deps.join(', ') + '], ' : '');

	var intro = 'define(' + params + 'function (' + args.join(', ') + ') { \'use strict\';\n\n';

	var exports = bundle.entryModule.exports;

	var exportBlock = undefined;

	if (exportMode === 'default') {
		exportBlock = 'return ' + bundle.entryModule.getCanonicalName('default') + ';';
	} else {
		exportBlock = '\n\n' + Object.keys(exports).map(function (name) {
			return 'exports.' + name + ' = ' + exports[name].localName + ';';
		}).join('\n');
	}

	return magicString.append(exportBlock).trim().indent().append('\n\n});').prepend(intro);
}

function cjs(bundle, magicString, exportMode) {
	var intro = '\'use strict\';\n\n';

	// TODO handle empty imports, once they're supported
	var importBlock = bundle.externalModules.map(function (module) {
		var requireStatement = 'var ' + module.name + ' = require(\'' + module.id + '\');';

		if (module.needsDefault) {
			requireStatement += '\n' + (module.needsNamed ? 'var ' + module.name + '__default = ' : '' + module.name + ' = ') + ('\'default\' in ' + module.name + ' ? ' + module.name + '[\'default\'] : ' + module.name + ';');
		}

		return requireStatement;
	}).join('\n');

	if (importBlock) {
		intro += importBlock + '\n\n';
	}

	magicString.prepend(intro);

	var exportBlock = undefined;
	if (exportMode === 'default' && bundle.entryModule.exports.default) {
		exportBlock = 'module.exports = ' + bundle.entryModule.getCanonicalName('default') + ';';
	} else if (exportMode === 'named') {
		exportBlock = keys(bundle.entryModule.exports).map(function (key) {
			var specifier = bundle.entryModule.exports[key];
			var name = bundle.entryModule.getCanonicalName(specifier.localName);

			return 'exports.' + key + ' = ' + name + ';';
		}).join('\n');
	}

	if (exportBlock) {
		magicString.append('\n\n' + exportBlock);
	}

	return magicString;
}

var es6__default = es6__es6;

function es6__es6(bundle, magicString, exportMode, options) {
	// TODO
	var introBlock = '';
	var exportBlock = '';

	return magicString.trim();
}

var iife = iife__es6;

function iife__es6(bundle, magicString, exportMode, options) {

	var intro = "(function () { 'use strict';\n\n";
	var outro = "\n\n})();";

	return magicString.trim().indent().prepend(intro).append(outro);
}

function umd(bundle, magicString, exportMode, options) {
	var indentStr = magicString.getIndentString();

	var globalNames = options.globals || {};

	var amdDeps = bundle.externalModules.map(quoteId);
	var cjsDeps = bundle.externalModules.map(req);
	var globalDeps = bundle.externalModules.map(function (module) {
		return has(globalNames, module.id) ? globalNames[module.id] : module.name;
	});

	var args = bundle.externalModules.map(getName);

	if (exportMode === 'named') {
		amdDeps.unshift('\'exports\'');
		cjsDeps.unshift('\'exports\'');
		globalDeps.unshift('(global.' + moduleName + ' = {})');

		args.unshift('exports');
	}

	var amdParams = (has(options, 'moduleId') ? '[\'' + options.moduleId + '\'], ' : '') + (amdDeps.length ? '[' + amdDeps.join(', ') + '], ' : '');

	var intro = ('(function (global, factory) {\n\t\t\ttypeof exports === \'object\' && typeof module !== \'undefined\' ? factory(' + cjsDeps.join(', ') + ') :\n\t\t\ttypeof define === \'function\' && define.amd ? define(' + amdParams + 'factory) :\n\t\t\tfactory(' + globalDeps + ');\n\t\t}(this, function (' + args + ') { \'use strict\';\n\n\t\t').replace(/^\t\t/gm, '').replace(/^\t/gm, indentStr);

	var exports = bundle.entryModule.exports;

	var exportBlock = '\n\n' + Object.keys(exports).map(function (name) {
		return 'exports.' + name + ' = ' + exports[name].localName + ';';
	}).join('\n');

	return magicString.append(exportBlock).trim().indent().append('\n\n}));').prepend(intro);
}

var finalisers = { amd: amd, cjs: cjs, es6: es6__default, iife: iife, umd: umd };

function replaceIdentifiers(statement, snippet, names) {
	var replacementStack = [names];

	var keys = Object.keys(names);

	if (keys.length === 0) {
		return;
	}

	walk(statement, {
		enter: function (node, parent) {
			var _this = this;

			var scope = node._scope;

			if (scope) {
				var _ret = (function () {
					var newNames = {};
					var hasReplacements = undefined;

					keys.forEach(function (key) {
						if (! ~scope.names.indexOf(key)) {
							newNames[key] = names[key];
							hasReplacements = true;
						}
					});

					if (!hasReplacements) {
						return {
							v: _this.skip()
						};
					}

					names = newNames;
					replacementStack.push(newNames);
				})();

				if (typeof _ret === 'object') return _ret.v;
			}

			// We want to rewrite identifiers (that aren't property names)
			if (node.type !== 'Identifier') return;
			if (parent.type === 'MemberExpression' && node !== parent.object) return;
			if (parent.type === 'Property' && node !== parent.value) return;
			// TODO others...?

			var name = has(names, node.name) && names[node.name];

			if (name && name !== node.name) {
				snippet.overwrite(node.start, node.end, name);
			}
		},

		leave: function (node) {
			if (node._scope) {
				replacementStack.pop();
				names = replacementStack[replacementStack.length - 1];
			}
		}
	});
}

function defaultResolver(importee, importer) {
	// absolute paths are left untouched
	if (_path.isAbsolute(importee)) return importee;

	// external modules stay external
	if (importee[0] !== '.') return false;

	return _path.resolve(_path.dirname(importer), importee).replace(/\.js$/, '') + '.js';
}

function Bundle___classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

function badExports(option, keys) {
	throw new Error('\'' + option + '\' was specified for options.exports, but entry module has following exports: ' + keys.join(', '));
}

var Bundle = (function () {
	function Bundle(options) {
		Bundle___classCallCheck(this, Bundle);

		this.entryPath = _path.resolve(options.entry).replace(/\.js$/, '') + '.js';
		this.base = _path.dirname(this.entryPath);

		this.resolvePath = options.resolvePath || defaultResolver;

		this.entryModule = null;
		this.modulePromises = {};
		this.statements = [];
		this.externalModules = [];
		this.defaultExportName = null;
		this.internalNamespaceModules = [];
	}

	Bundle.prototype.fetchModule = function fetchModule(importee, importer) {
		var _this = this;

		return sander.Promise.resolve(importer === null ? importee : this.resolvePath(importee, importer)).then(function (path) {
			if (!path) {
				// external module
				if (!has(_this.modulePromises, importee)) {
					var _module = new ExternalModule(importee);
					_this.externalModules.push(_module);
					_this.modulePromises[importee] = sander.Promise.resolve(_module);
				}

				return _this.modulePromises[importee];
			}

			if (!has(_this.modulePromises, path)) {
				_this.modulePromises[path] = sander.readFile(path, { encoding: 'utf-8' }).then(function (code) {
					var module = new Module({
						path: path,
						code: code,
						bundle: _this
					});

					return module;
				});
			}

			return _this.modulePromises[path];
		});
	};

	Bundle.prototype.build = function build() {
		var _this2 = this;

		// bring in top-level AST nodes from the entry module
		return this.fetchModule(this.entryPath, null).then(function (entryModule) {
			_this2.entryModule = entryModule;

			if (entryModule.exports.default) {
				var defaultExportName = makeLegalIdentifier(_path.basename(_this2.entryPath).slice(0, -_path.extname(_this2.entryPath).length));
				while (entryModule.ast._scope.contains(defaultExportName)) {
					defaultExportName = '_' + defaultExportName;
				}

				entryModule.suggestName('default', defaultExportName);
			}

			return entryModule.expandAllStatements(true);
		}).then(function (statements) {
			_this2.statements = statements;
			_this2.deconflict();
		});
	};

	Bundle.prototype.deconflict = function deconflict() {
		var definers = {};
		var conflicts = {};

		// Discover conflicts (i.e. two statements in separate modules both define `foo`)
		this.statements.forEach(function (statement) {
			keys(statement._defines).forEach(function (name) {
				if (has(definers, name)) {
					conflicts[name] = true;
				} else {
					definers[name] = [];
				}

				// TODO in good js, there shouldn't be duplicate definitions
				// per module... but some people write bad js
				definers[name].push(statement._module);
			});
		});

		// Assign names to external modules
		this.externalModules.forEach(function (module) {
			var name = makeLegalIdentifier(module.id);

			while (has(definers, name)) {
				name = '_' + name;
			}

			module.name = name;
		});

		// Rename conflicting identifiers so they can live in the same scope
		keys(conflicts).forEach(function (name) {
			var modules = definers[name];

			modules.pop(); // the module closest to the entryModule gets away with keeping things as they are

			modules.forEach(function (module) {
				module.rename(name, name + '$' + ~ ~(Math.random() * 100000)); // TODO proper deconfliction mechanism
			});
		});
	};

	Bundle.prototype.generate = function generate() {
		var options = arguments[0] === undefined ? {} : arguments[0];

		var magicString = new MagicString.Bundle({ separator: '' });

		// Determine export mode - 'default', 'named', 'none'
		var exportMode = this.getExportMode(options.exports);

		var previousMargin = 0;

		// Apply new names and add to the output bundle
		this.statements.forEach(function (statement) {
			var replacements = {};

			keys(statement._dependsOn).concat(keys(statement._defines)).forEach(function (name) {
				var canonicalName = statement._module.getCanonicalName(name);

				if (name !== canonicalName) {
					replacements[name] = canonicalName;
				}
			});

			var source = statement._source.clone().trim();

			// modify exports as necessary
			if (/^Export/.test(statement.type)) {
				// skip `export { foo, bar, baz }`
				if (statement.type === 'ExportNamedDeclaration' && statement.specifiers.length) {
					return;
				}

				// remove `export` from `export var foo = 42`
				if (statement.type === 'ExportNamedDeclaration' && statement.declaration.type === 'VariableDeclaration') {
					source.remove(statement.start, statement.declaration.start);
				}

				// remove `export` from `export class Foo {...}` or `export default Foo`
				// TODO default exports need different treatment
				else if (statement.declaration.id) {
					source.remove(statement.start, statement.declaration.start);
				}

				// declare variables for expressions
				else {
					var _name = statement.type === 'ExportDefaultDeclaration' ? 'default' : 'TODO';
					var canonicalName = statement._module.getCanonicalName(_name);
					source.overwrite(statement.start, statement.declaration.start, 'var ' + canonicalName + ' = ');
				}
			}

			replaceIdentifiers(statement, source, replacements);

			// add leading comments
			if (statement._leadingComments.length) {
				var commentBlock = statement._leadingComments.map(function (comment) {
					return comment.block ? '/*' + comment.text + '*/' : '//' + comment.text;
				}).join('\n');

				magicString.addSource(new MagicString(commentBlock));
			}

			// add margin
			var margin = Math.max(statement._margin[0], previousMargin);
			var newLines = new Array(margin).join('\n');

			// add the statement itself
			magicString.addSource({
				content: source,
				separator: newLines
			});

			// add trailing comments
			var comment = statement._trailingComment;
			if (comment) {
				var commentBlock = comment.block ? ' /*' + comment.text + '*/' : ' //' + comment.text;

				magicString.append(commentBlock);
			}

			previousMargin = statement._margin[1];
		});

		// prepend bundle with internal namespaces
		var indentString = magicString.getIndentString();
		var namespaceBlock = this.internalNamespaceModules.map(function (module) {
			var exportKeys = keys(module.exports);

			return 'var ' + module.getCanonicalName('*') + ' = {\n' + exportKeys.map(function (key) {
				return '' + indentString + 'get ' + key + ' () { return ' + module.getCanonicalName(key) + '; }';
			}).join(',\n') + '\n};\n\n';
		}).join('');

		magicString.prepend(namespaceBlock);

		var finalise = finalisers[options.format || 'es6'];

		if (!finalise) {
			throw new Error('You must specify an output type - valid options are ' + keys(finalisers).join(', '));
		}

		magicString = finalise(this, magicString.trim(), exportMode, options);

		return {
			code: magicString.toString(),
			map: magicString.generateMap({
				includeContent: true,
				file: options.dest
				// TODO
			})
		};
	};

	Bundle.prototype.getExportMode = function getExportMode(exportMode) {
		var exportKeys = keys(this.entryModule.exports);

		if (exportMode === 'default') {
			if (exportKeys.length !== 1 || exportKeys[0] !== 'default') {
				badExports('default', exportKeys);
			}
		} else if (exportMode === 'none' && exportKeys.length) {
			badExports('none', exportKeys);
		}

		if (!exportMode || exportMode === 'auto') {
			if (exportKeys.length === 0) {
				exportMode = 'none';
			} else if (exportKeys.length === 1 && exportKeys[0] === 'default') {
				exportMode = 'default';
			} else {
				exportMode = 'named';
			}
		}

		if (!/(?:default|named|none)/.test(exportMode)) {
			throw new Error('options.exports must be \'default\', \'named\', \'none\', \'auto\', or left unspecified (defaults to \'auto\')');
		}

		return exportMode;
	};

	return Bundle;
})();

var SOURCEMAPPING_URL = 'sourceMa';
SOURCEMAPPING_URL += 'ppingURL';
function rollup(entry) {
	var options = arguments[1] === undefined ? {} : arguments[1];

	var bundle = new Bundle({
		entry: entry,
		resolvePath: options.resolvePath
	});

	return bundle.build().then(function () {
		return {
			generate: function (options) {
				return bundle.generate(options);
			},
			write: function (dest) {
				var options = arguments[1] === undefined ? {} : arguments[1];

				var _bundle$generate = bundle.generate({
					dest: dest,
					format: options.format,
					globalName: options.globalName
				});

				var code = _bundle$generate.code;
				var map = _bundle$generate.map;

				code += '\n//# ' + SOURCEMAPPING_URL + '=' + _path.basename(dest) + '.map';

				return Promise.all([sander.writeFile(dest, code), sander.writeFile(dest + '.map', map.toString())]);
			}
		};
	});
}

exports.rollup = rollup;
//# sourceMappingURL=rollup.js.map
